# -*- coding: utf-8 -*-
__author__ = 'Monica Figuera'

import time

from TravSHACL.rule_based_validation.InstancesRetrieval import InstancesRetrieval
from TravSHACL.utils import fileManagement
from TravSHACL.utils.ValidationStats import ValidationStats


class Validation:
    """This class is responsible for managing the validation process."""

    def __init__(self, endpoint_url, node_order, shapes_dict, target_shape_predicates, use_selective_queries,
                 output_dir_name, save_stats, save_targets_to_file):
        """
        Creates a new instance for the validation process.

        :param endpoint_url: URL of the SPARQL endpoint to evaluate
        :param node_order: indicates the order in which the shapes will be evaluated
        :param shapes_dict: a Python dictionary holding all shapes of the shape schema
        :param target_shape_predicates: names of the shapes with a target definition
        :param use_selective_queries: indicates whether selective queries will be used
        :param output_dir_name: path for the output files
        :param save_stats: indicates whether statistics will be saved to the output path
        :param save_targets_to_file: indicates whether target classifications will be saved to the output path
        """
        self.node_order = node_order
        self.shapes_dict = shapes_dict
        self.target_shape_predicates = target_shape_predicates
        self.selectivity_enabled = use_selective_queries

        self.output_dir_name = output_dir_name
        self.save_stats = save_stats
        self.save_targets_to_file = save_targets_to_file
        self.stats = ValidationStats()
        self.traces = set()

        self.stats.update_log('Node order: ' + str(self.node_order) + '\n')
        self.InstRetrieval = InstancesRetrieval(endpoint_url, shapes_dict, self.stats)
        self.valid_targets_after_termination = set()
        self.start_of_verification = time.time() * 1000.0

    def exec(self):
        """Executes the validation process of the entire shape schema."""
        focus_shape = self.shapes_dict[self.node_order.pop(0)]
        state = ValidationState(self.shapes_dict)
        initial_targets = self.retrieve_next_targets(state, focus_shape, state.shapes_state)
        state.remaining_targets.update(initial_targets)
        state.shapes_state[focus_shape.get_id()]['remaining_targets_count'] = len(initial_targets)

        start = time.time() * 1000.0
        self.validate(state, focus_shape)
        finish = time.time() * 1000.0
        elapsed = round(finish - start)
        self.stats.record_total_time(elapsed)
        print('Total execution time: ', str(elapsed), ' ms')
        self.stats.record_total_rules(state.total_rule_number)
        self.stats.update_log('\n\nMaximal number or rules in memory: ' + str(self.stats.max_rules))
        self.stats.update_log('\nTotal number of rules: ' + str(state.total_rule_number))
        return self.validation_output(state.shapes_state)

    def validate(self, state, focus_shape):
        """
        Validates a shape.

        :param state: current validation state
        :param focus_shape: focus shape to be evaluated
        """
        if len(state.visited_shapes) == len(self.shapes_dict) or focus_shape is None:
            self.valid_targets_after_termination.update(state.remaining_targets)
            return

        self.stats.update_log('\n\n>>>>> Starting validation of shape: ' + focus_shape.get_id())
        state.evaluated_predicates.add(focus_shape.get_id())
        self.eval_shape(state, focus_shape, state.shapes_state)

        next_focus_shape = None
        if len(self.node_order) > 0:
            next_focus_shape = self.shapes_dict[self.node_order.pop(0)]
            pending_targets = self.retrieve_next_targets(state, next_focus_shape, state.shapes_state)
            state.remaining_targets.update(pending_targets)

        self.validate(state, next_focus_shape)

    def retrieve_next_targets(self, state, next_focus_shape, shapes_state):
        """
        Runs query to get targets of the next focus shape. If the query was rewritten (filtering), pending_targets are
        added to set of remaining targets and invalid_targets are directly classified.

        :param state: current overall validation state
        :param next_focus_shape: the upcoming focus shape, i.e., the shape all possible targets need to be collected for
        :param shapes_state: dictionary storing validation's state of each shape
        :return: all pending targets for the upcoming focus shape
        """
        self.stats.update_log('\n\n>>>>>\nRetrieving (next) targets ...')
        if next_focus_shape.get_target_query() is None:
            return set()

        next_focus_shape_name = next_focus_shape.get_id()
        filtering_shape = self.get_evaluated_out_neighbor(next_focus_shape, state.visited_shapes, state)
        shapes_state[next_focus_shape_name]['filtering_shape'] = filtering_shape
        max_ref = next_focus_shape.is_max_ref(filtering_shape.get_id()) if filtering_shape is not None else True  # TODO: check if filtering based on max constraints is possible
        if self.selectivity_enabled and filtering_shape is not None and not max_ref:
            # A query can filter its answers with (in)validated targets given by an evaluated out-neighboring shape
            pending, invalid = self.InstRetrieval.extract_targets_with_filter(next_focus_shape, filtering_shape)
            for target in invalid:
                self.register_target(target, 'violated', next_focus_shape_name, shapes_state)
                shapes_state[next_focus_shape_name]['inferred'].add((target[0], target[1], not target[2]))
                shapes_state[next_focus_shape_name]['remaining_targets_count'] -= 1
        else:
            pending = self.InstRetrieval.extract_targets(next_focus_shape)

        shapes_state[next_focus_shape_name]['remaining_targets_count'] = len(pending)
        return pending

    @staticmethod
    def get_evaluated_out_neighbor(focus_shape, visited_shapes, state):
        """
        Returns the name of an outgoing neighboring shape (i.e., object in inter-shape constraint of 'focus_shape')
        that was already evaluated and contains relevant information for query filtering:
            - the number of valid or invalid targets is less than a threshold (currently set as 256),
            - the list of invalid targets is not empty,
            - the neighboring shape has a target query defined.

        :param focus_shape: current shape being evaluated
        :param visited_shapes: set with names of shapes that were already evaluated
        :param state: object keeping track of the current validation state
        :return: reference to the (best) previously evaluated shape, or None if no shape fulfills the conditions
        """
        best_filtering_shape = None
        best_val_threshold = 256
        best_inv_threshold = 256
        for prev_shape in visited_shapes:
            prev_shape_name = prev_shape.get_id()
            if focus_shape.referencedShapes.get(prev_shape_name) is not None:
                len_val = len(prev_shape.get_valid_targets())
                len_inv = len(prev_shape.get_invalid_targets())

                if (0 < len_val < best_val_threshold) or (0 < len_inv < best_inv_threshold):
                    if len_inv > 0 and prev_shape.get_target_query() is not None:
                        if state.shapes_state[prev_shape_name]['remaining_targets_count'] == 0:
                            best_filtering_shape = prev_shape

        return best_filtering_shape

    def eval_shape(self, state, shape, shapes_state):
        """
        Interleaves 'shape' entities and performs a deferred saturation for entities that could not be validated during
        the interleaving because of missing information to be provided by (not yet evaluated) out-neighboring shapes.

        :param state: object keeping track of the current validation state
        :param shape: focus shape to be evaluated
        :param shapes_state: dictionary storing validation's state of each shape
        """
        shape_name = shape.get_id()

        self.eval_constraints_queries(state, shape, shapes_state[shape_name]['filtering_shape'])  # interleave

        self.stats.update_log('\nStarting saturation ...')
        start = time.time() * 1000.0
        self.saturate_remaining(state, shape_name, shapes_state)
        end = time.time() * 1000.0

        self.stats.record_saturation_time(end - start)
        self.stats.update_log('\nsaturation ...\nelapsed: ' + str(end - start) + ' ms')

        state.visited_shapes.add(shape)
        self.stats.update_log('\nRemaining targets: ' + str(len(state.remaining_targets)))

    def eval_constraints_queries(self, state, shape, filtering_shape):
        """
        Evaluates all constraints queries of focus 'shape'.
        There is one query with all min constraints and (possibly) one query per max constraint.

        :param state: object keeping track of the current validation state
        :param shape: focus shape to be evaluated
        :param filtering_shape: shape (referenced by focus 'shape') providing filtering instances
        """
        shape_rp = shape.get_rule_pattern()
        shape_name = shape.get_id()
        shapes_state = state.shapes_state
        remaining_targets = state.remaining_targets

        if shape.minQuery is None and not shape.maxQueries:  # current shape is a shape without constraints
            to_remove = []
            for head in remaining_targets:
                if head[0] == shape_name:
                    self.register_target(head, 'valid', shape_name, shapes_state)
                    to_remove.append(head)
                    shapes_state[shape_name]['remaining_targets_count'] -= 1
                    shapes_state[shape_name]['inferred'].add(head)
            for head in to_remove:
                remaining_targets.discard(head)

        if shape.minQuery is not None:
            min_query_rp = shape.minQuery.get_rule_pattern()
            self.interleave(state, shape, shape.minQuery, filtering_shape, min_query_rp, shape_rp, 'min')

        for q in shape.maxQueries:
            max_query_rp = q.get_rule_pattern()
            self.interleave(state, shape, q, filtering_shape, max_query_rp, shape_rp, 'max')
            # after running the max constraints, rules need to be added for the instances that were not
            # in the query result since they will still be valid and may need to be checked further
            for head in remaining_targets:
                if head[0] == shape_name:
                    body = set()
                    for i, atom_pattern in enumerate(shape_rp.body):
                        a = (atom_pattern[0], head[1], atom_pattern[2])
                        body.add(a)

                    s_head = (shape_name, head[1], True)

                    if s_head not in state.rule_map.keys():
                        s = set()
                        s.add(frozenset(body))
                        state.rule_map[s_head] = s
                        state.rule_number += 1
                        state.total_rule_number += 1
                    else:
                        if frozenset(body) not in state.rule_map[s_head]:
                            state.rule_number += 1
                            state.total_rule_number += 1
                        state.rule_map[s_head].add(frozenset(body))

    def interleave(self, state, shape, q, filtering_shape, q_rule_pattern, s_rule_pattern, q_type):
        """
        Evaluates corresponding min/max query of 'shape' and starts interleaving process for all answers retrieved.

        Query 'q' can be rewritten to include a filter (VALUES clause) with the targets of 'filtering_shape', and
        (possibly) partitioned when the number of filtering instances exceeds a given threshold.

        Each grounded rule is composed of literals (an atom or its negation).
        Each atom 'a' is a tuple with 3 fields: a[0] - predicate, a[1] - instance (entity), a[2] - sign (True or False)

        :param state: current state of the validation
        :param shape: focus shape
        :param q: constraint query to be evaluated
        :param filtering_shape: shape (referenced by focus 'shape') providing filtering instances
        :param q_rule_pattern: query rule pattern
        :param s_rule_pattern: shape rule pattern
        :param q_type: query type ('min' or 'max')
        """
        shapes_state = state.shapes_state
        preds_to_shapes = state.preds_to_shapes
        shape_name = shape.get_id()
        t_state = shapes_state[shape_name]  # focus shape's state

        query_rp_head = q_rule_pattern.head
        query_rp_body = q_rule_pattern.body
        shape_rp_head = s_rule_pattern.head
        shape_rp_body = s_rule_pattern.body

        q_body_ref_shapes = [preds_to_shapes[pattern[0]] for pattern in query_rp_body]
        s_body_ref_shapes = [preds_to_shapes[pattern[0]] for pattern in shape_rp_body]

        shape_max_refs = shape.get_max_query_valid_refs()
        inter_constr_count = {}
        new_rules_count = 0
        rules_directly_inferred = 0

        for query_str in self.InstRetrieval.rewrite_constraint_query(shape, q, filtering_shape, q_type, self.selectivity_enabled):
            start = time.time() * 1000.0
            for b in self.InstRetrieval.run_constraint_query(q, query_str):
                q_head = (query_rp_head[0], b[query_rp_head[1]]['value'], query_rp_head[2])
                s_head = (shape_rp_head[0], b[shape_rp_head[1]]['value'], shape_rp_head[2])

                body = set()
                is_body_inferred = True
                is_body_inferrable = True
                negated_body = False
                for i, atom_pattern in enumerate(query_rp_body):
                    a_state = shapes_state[q_body_ref_shapes[i]]  # body atom's shape state
                    a = (atom_pattern[0], b[atom_pattern[1]]['value'], atom_pattern[2])
                    body.add(a)
                    if a[0] in state.evaluated_predicates:  # exclude not yet evaluated atom's query
                        if state.rule_map.get((a[0], a[1], True)) is None:
                            if a not in a_state['inferred']:
                                if (a[0], a[1], not a[2]) not in a_state['inferred']:
                                    is_body_inferred = False
                                    continue  # exclude non-selective answers from interleaving
                                else:
                                    is_body_inferred = False
                                    is_body_inferrable = False
                        else:
                            is_body_inferred = False

                        # if interleaving min-query with an upper bound <> None -> verify upper bound violations
                        if a in a_state['inferred'] \
                                and q_type == 'min' \
                                and shape_max_refs.get(q_body_ref_shapes[i]) is not None:
                            if inter_constr_count.get(s_head) is None:
                                inter_constr_count[s_head] = {inter_constraint: set() for inter_constraint in shape_max_refs}
                            inter_constr_count[s_head][q_body_ref_shapes[i]].add(a)
                            if len(inter_constr_count[s_head][q_body_ref_shapes[i]]) > shape_max_refs[q_body_ref_shapes[i]]:
                                negated_body = True  # if number of instantiations exceeds max value -> infer negation
                    else:
                        is_body_inferred = False

                if negated_body:
                    # case (2) - infer negation of rule head (given negated body atom)
                    t_state['inferred'].add((q_head[0], q_head[1], False))
                    if s_head[2] and s_head in state.remaining_targets:
                        self.register_target(s_head, 'violated', shape_name, shapes_state)
                        state.remaining_targets.discard(s_head)
                        t_state['remaining_targets_count'] -= 1
                    continue

                # case (3) - add pending rule / infer query head
                if not is_body_inferred:
                    if is_body_inferrable:
                        if q_head not in state.rule_map.keys():
                            s = set()
                            s.add(frozenset(body))
                            state.rule_map[q_head] = s
                            new_rules_count += 1
                        else:
                            if frozenset(body) not in state.rule_map[q_head]:
                                new_rules_count += 1
                                state.rule_map[q_head].add(frozenset(body))
                else:
                    t_state['inferred'].add(q_head)
                    rules_directly_inferred += 1

                # Shape rule pattern #
                body = set()
                is_body_inferred = True
                negated_body = False
                for i, atom_pattern in enumerate(shape_rp_body):
                    a_state = shapes_state[s_body_ref_shapes[i]]
                    a = (atom_pattern[0], b[atom_pattern[1]]['value'], atom_pattern[2])
                    body.add(a)
                    if a not in a_state['inferred']:
                        if (a[0], a[1], not a[2]) not in a_state['inferred']:
                            is_body_inferred = False
                            continue  # non-selective (when retrieved objects of a triple that do not match any target)

                        # case (1) - if negated (unmatchable) body atoms
                        elif (a[0], a[1], not a[2]) in a_state['inferred']:
                            negated_body = True
                            break
                        else:
                            is_body_inferred = False

                if negated_body:
                    # case (2) - infer negation of rule head (negated body atom)
                    t_state['inferred'].add((s_head[0], s_head[1], False))
                    if s_head[2] and s_head in state.remaining_targets:
                        self.register_target(s_head, 'violated', shape_name, shapes_state)
                        state.remaining_targets.discard(s_head)
                        t_state['remaining_targets_count'] -= 1
                    continue

                # case (3) - add pending rule / classify valid target (all body inferred)
                if not is_body_inferred:
                    if s_head not in state.rule_map.keys():
                        s = set()
                        s.add(frozenset(body))
                        state.rule_map[s_head] = s
                        new_rules_count += 1
                    else:
                        if frozenset(body) not in state.rule_map[s_head]:
                            new_rules_count += 1
                            state.rule_map[s_head].add(frozenset(body))
                else:
                    if s_head not in t_state['inferred']:
                        t_state['inferred'].add(s_head)
                    if s_head in state.remaining_targets:
                        for a in body:
                            t_state['inferred'].discard(a)  # de-allocating memory

                        self.register_target(s_head, 'valid', shape_name, shapes_state)
                        state.remaining_targets.discard(s_head)
                        t_state['remaining_targets_count'] -= 1
                    rules_directly_inferred += 1

            self.stats.update_log('\nGrounded rules. \n')
            end = time.time()*1000.0
            self.stats.record_interleaving_time(end - start)
            state.evaluated_predicates.add(query_rp_head[0])

        all_current_rules = new_rules_count + rules_directly_inferred
        state.rule_number += new_rules_count
        state.total_rule_number += all_current_rules
        self.stats.update_log('\n\nNumber of rules: ' + str(all_current_rules))
        self.stats.record_current_number_of_rules(all_current_rules)

    def saturate_remaining(self, state, shape_name, shapes_state):
        """
        The saturation process consists of two steps:
        1. Negate: same as in step (1) of interleaving process
        2. Infer: same is step (2) of interleaving process
        Repeat 1 and 2 until no further changes are made (i.e., no new inferences found)

        :param state: current state of the validation
        :param shape_name: name of focus shape or its 'parents' (after recursion)
        :param shapes_state: dictionary storing validation's state of each shape
        """
        rule_map = state.rule_map
        negated = self.negate_unmatchable_heads(state, rule_map, state.evaluated_predicates, shape_name, shapes_state, state.preds_to_shapes)
        inferred = self.apply_rules(state, state.remaining_targets, rule_map, shape_name, shapes_state, state.preds_to_shapes)
        if negated or inferred:
            self.saturate_remaining(state, shape_name, shapes_state)

    def negate_unmatchable_heads(self, state, rule_map, evaluated_predicates, shape_name, shapes_state, preds_to_shapes):
        """
        Derives negation of atoms that are not satisfied. An atom 'a' is not satisfied when:
        its query (predicate) was already evaluated, positive 'a' is not a rule head, and 'a' is not inferred yet.

        :param state: current state of the validation
        :param rule_map: set of pending rules associated to either focus shape or an incoming neighbor after recursion
        :param evaluated_predicates: target/min/max query predicates evaluated so far
        :param shape_name: string name of the focus shape
        :param shapes_state: dictionary storing validation's state of each shape
        :param preds_to_shapes: dictionary that maps predicate names to shape names
        :return: True if new negative inferences were found, False otherwise
        """
        new_negated_atom_found = False
        all_body_atoms = set(frozenset().union(*set().union(*rule_map.values())))
        for a in all_body_atoms:
            a_state = shapes_state[preds_to_shapes[a[0]]]  # atom's shape state (gets shape id from the constraint id)
            if a[0] in evaluated_predicates \
                    and rule_map.get((a[0], a[1], True)) is None \
                    and a not in a_state['inferred']:  # if atom 'a' is not satisfied
                negated_atom = (a[0], a[1], False)
                if negated_atom not in a_state['inferred']:
                    new_negated_atom_found = True
                    a_state['inferred'].add(negated_atom)

        remaining = set()
        for a in state.remaining_targets:
            t_state = shapes_state[a[0]]  # target's shape state
            if a[0] in evaluated_predicates \
                    and rule_map.get((a[0], a[1], True)) is None \
                    and a not in t_state['inferred']:  # if atom 'a' is not satisfied
                self.register_target(a, 'violated', shape_name, shapes_state)
                t_state['inferred'].add((a[0], a[1], not a[2]))
                t_state['remaining_targets_count'] -= 1
            else:
                remaining.add(a)
        state.remaining_targets = remaining
        return new_negated_atom_found

    def apply_rules(self, state, remaining_targets, rule_map, shape_name, shapes_state, preds_to_shapes):
        """
        Performs two types of inferences:
        # case (1): If the rule map contains a rule and some rule bodies were inferred
                    => head of the rule is inferred, rule dropped.
        # case (2): If the negation of any rule body was inferred
                    => the rule cannot be applied (rule head not inferred), rule dropped.

        :param state: current state of the validation
        :param remaining_targets: pending targets, i.e., targets that are neither valid nor invalid yet
        :param rule_map: set of pending rules associated to either focus shape or an incoming neighbor after recursion
        :param shape_name: name of the current focus shape
        :param shapes_state: dictionary storing validation's state of each shape
        :param preds_to_shapes: dictionary that maps predicate names to shape names
        :return: True if new inferences were found, False otherwise
        """
        fresh_literals = False
        rule_map_copy = rule_map.copy()
        for head, bodies in rule_map_copy.items():
            head_shape_name = preds_to_shapes[head[0]]
            inferred_bodies = set()
            for body in bodies:
                inferred_atoms = {'F_atom' if (a[0], a[1], not a[2]) in shapes_state[preds_to_shapes[a[0]]]['inferred']
                                  else ('T_atom' if a in shapes_state[preds_to_shapes[a[0]]]['inferred']
                                        else 'P_atom') for a in body}
                if 'T_atom' in inferred_atoms and len(inferred_atoms) == 1:
                    inferred_bodies.add('T')
                elif 'F_atom' in inferred_atoms:
                    inferred_bodies.add('F')
                else:
                    inferred_bodies.add('P')

            if 'T' in inferred_bodies:  # case (1)
                fresh_literals = True
                if head in remaining_targets:
                    self.register_target(head, 'valid', shape_name, shapes_state)
                    remaining_targets.discard(head)
                    shapes_state[head_shape_name]['remaining_targets_count'] -= 1
                shapes_state[head_shape_name]['inferred'].add(head)
                del rule_map[head]
                state.rule_number -= len(bodies)
            elif 'F' in inferred_bodies and 'P' not in inferred_bodies:  # case (2)
                fresh_literals = True
                if head in remaining_targets:
                    self.register_target(head, 'violated', shape_name, shapes_state)
                    remaining_targets.discard(head)
                    shapes_state[head_shape_name]['remaining_targets_count'] -= 1
                shapes_state[head_shape_name]['inferred'].add((head[0], head[1], not head[2]))
                del rule_map[head]
                state.rule_number -= len(bodies)

        if not fresh_literals:
            return False
        self.stats.update_log('\nRemaining targets: ' + str(len(remaining_targets)))
        return True

    def register_target(self, t, t_type, invalidating_shape_name, shapes_state):
        """
        Adds each target to the set of valid/invalid instances of a shape.

        :param t: target tuple: t[0] - predicate (string), t[1] - instance (string), t[2] - sign (boolean)
        :param t_type: string value that tells whether 't' is a valid target ('valid') or a 'violated' one
        :param invalidating_shape_name: string name of the shape that (in)validates target 't'
        :param shapes_state: dictionary storing validation's state of each shape
        """
        instance = '<' + t[1] + '>'
        self.shapes_dict[t[0]].targets[t_type].add(instance)
        shapes_state[invalidating_shape_name]['registered_targets'][t_type].add(t)
        self.traces.add(''.join([t_type, ', ',
                        str(len(self.traces)), ', ',
                        str(round(time.time() * 1000.0 - self.start_of_verification)), '\n']))

    @staticmethod
    def write_targets_to_file(output_dir_name, all_valid_targets, all_invalid_targets):
        """Writes all target classifications to file, i.e., one file for valid targets and one for invalid ones."""
        def write_list_to_file(list_, file_):
            """Helper function to write a list to file."""
            for elem in list_:
                lit_sign_str = '' if elem[2] else '!'
                lit_str = lit_sign_str + elem[0] + '(' + elem[1] + ')'  # get string representation of tuple
                file_.write(lit_str + ',\n')

        valid_targets_file = fileManagement.open_file(output_dir_name, 'targets_valid.log')
        write_list_to_file(all_valid_targets, valid_targets_file)
        fileManagement.close_file(valid_targets_file)

        violated_targets_file = fileManagement.open_file(output_dir_name, 'targets_violated.log')
        write_list_to_file(all_invalid_targets, violated_targets_file)
        fileManagement.close_file(violated_targets_file)

    def validation_output(self, shapes_state):
        """
        Saves to local file the (in)validated targets and returns result of validation

        :param shapes_state: dictionary storing validation's state of each shape
        :return: dictionary containing shape names and their respective (in)validated targets
        """
        output = {}
        all_valid_targets = set()
        all_invalid_targets = set()
        for shape_name in self.shapes_dict.keys():
            validated_targets = shapes_state[shape_name]['registered_targets']['valid']
            invalidated_targets = shapes_state[shape_name]['registered_targets']['violated']
            output[shape_name] = {
                'valid_instances': validated_targets,
                'invalid_instances': invalidated_targets,
            }
            all_valid_targets.update(validated_targets)
            all_invalid_targets.update(invalidated_targets)
        all_valid_targets.update(self.valid_targets_after_termination)

        if self.save_targets_to_file:
            self.write_targets_to_file(self.output_dir_name, all_valid_targets, all_invalid_targets)

        if self.save_stats:
            validation_log = fileManagement.open_file(self.output_dir_name, 'validation.log')
            stats = fileManagement.open_file(self.output_dir_name, 'stats.txt')
            traces = fileManagement.open_file(self.output_dir_name, 'traces.csv')

            self.stats.record_number_of_targets(len(all_valid_targets), len(all_invalid_targets))
            self.stats.write_all_stats(stats)
            fileManagement.close_file(stats)

            self.stats.write_validation_log(validation_log)
            validation_log.write('\nValid targets: ' + str(len(all_valid_targets)))
            validation_log.write('\nInvalid targets: ' + str(len(all_invalid_targets)))
            fileManagement.close_file(validation_log)

            traces.write('Target, #TripleInThatNode, TimeSinceStartOfVerification(ms)\n')
            for trace in self.traces:
                traces.write(trace)
            fileManagement.close_file(traces)

        # add to output all targets that could not be (in)validated by any shape
        output["unbound"] = {'valid_instances': self.valid_targets_after_termination}

        # TTL validation report
        if self.save_stats:
            if len(all_invalid_targets) == 0:
                output_ttl = ':report a sh:ValidationReport ;\n' + '  sh:conforms true '
            else:
                output_ttl = ':report a sh:ValidationReport ;\n' + '  sh:conforms false ;\n' + '  sh:result'
                for i, violation in enumerate(all_invalid_targets):
                    if i != 0:
                        output_ttl += ' ,'
                    output_ttl += '\n    [ a  sh:ValidationResult ;\n' +\
                                  '      sh:resultSeverity  sh:Violation ;\n' +\
                                  '      sh:focusNode  <' + violation[1] + '> ;\n' +\
                                  '      sh:sourceShape  ' + violation[0] + ' ]'

            output_ttl = '@prefix sh: <http://www.w3.org/ns/shacl#> . \n\n' + output_ttl + ' .'

            validation_report = fileManagement.open_file(self.output_dir_name, 'validationReport.ttl')
            validation_report.write(output_ttl)
            fileManagement.close_file(validation_report)

        return output


class ValidationState:
    """This class is responsible for keeping track of the validation state."""

    def __init__(self, shapes_dict):
        self.remaining_targets = set()
        self.visited_shapes = set()
        self.evaluated_predicates = set()
        self.shapes_state = {}
        self.preds_to_shapes = {}  # maps all constraint ids to their respective shape ids
        self.rule_map = {}
        self.rule_number = 0
        self.total_rule_number = 0

        for shape_name in shapes_dict.keys():
            self.shapes_state[shape_name] = {
                'filtering_shape': None,
                'inferred': set(),
                'remaining_targets_count': 0,
                'registered_targets': {'valid': set(), 'violated': set()}
            }
            for pred in shapes_dict[shape_name].predicates:
                self.preds_to_shapes[pred] = shape_name



# -*- coding: utf-8 -*-
__author__ = 'Monica Figuera'

import math
import time

from TravSHACL.sparql.SPARQLEndpoint import SPARQLEndpoint
from TravSHACL.sparql.QueryGenerator import get_target_node_statement
from TravSHACL.constraints.MinOnlyConstraint import MinOnlyConstraint
from TravSHACL.constraints.MaxOnlyConstraint import MaxOnlyConstraint


class InstancesRetrieval:
    """This class is responsible for retrieving the instances from a SPARQL endpoint."""

    def __init__(self, endpoint_url, shapes_dict, stats):
        """
        Creates a new instance for the data retrieval.

        :param endpoint_url: URL of the SPARQL endpoint to collect the data from
        :param shapes_dict: a Python dictionary holding all shapes of the shape schema
        :param stats: instance of ValidationStats to keep the statistics up-to-date
        """
        self.endpointURL = endpoint_url
        self.endpoint = SPARQLEndpoint(endpoint_url)
        self.shapes_dict = shapes_dict
        self.stats = stats

    def run_constraint_query(self, q, query_str):
        """
        Retrieves answers from the SPARQL endpoint.

        :param q: Query object representing the SPARQL query to answer
        :param query_str: possibly rewritten query string, e.g., a partition of the original query
        :return: answer set bindings for the constraint query
        """
        self.stats.update_log(''.join(['\n\nEvaluating query ', q.get_id(), ':\n', query_str]))

        start = time.time()*1000.0
        bindings = self.endpoint.run_query(query_str)['results']['bindings']
        end = time.time()*1000.0

        self.stats.update_log(''.join(['\nelapsed: ', str(end - start), ' ms\n']))
        self.stats.record_query_exec_time(end - start)
        self.stats.record_query()
        self.stats.update_log(''.join(['\nNumber of solution mappings: ', str(len(bindings)), '\n']))
        self.stats.record_number_of_sol_mappings(len(bindings))
        return bindings

    def extract_targets(self, shape):
        """
        Retrieves answers from the SPARQL endpoint.
        When a network of shapes has only 'some' targets, a shape without a target class returns no new bindings

        :param shape: focus shape being evaluated
        :return: set containing target literals (stored in the form of built-in python tuples)
        """
        query = shape.get_target_query()  # targetQuery is set in shape's definition file (json file)
        self.stats.update_log(''.join(['\nEvaluating target query for ', shape.id, ':\n', query]))

        start = time.time() * 1000.0
        bindings = self.endpoint.run_query(query)['results']['bindings']
        end = time.time() * 1000.0

        self.stats.update_log('\nelapsed: ' + str(end - start) + ' ms\n')
        self.stats.record_query_exec_time(end - start)
        self.stats.record_query()
        self.stats.update_log('\nNumber of targets retrieved: ' + str(len(bindings)))
        return {(shape.id, b['x']['value'], True) for b in bindings}

    def extract_targets_with_filter(self, shape, filtering_shape):
        """
        Retrieves more selective query answers by separating early invalidated targets from the still pending ones.

        :param shape: focus shape being evaluated
        :param filtering_shape: referenced shape used to filter query, or None
        :return: two sets containing all targets of 'shape': pending targets to validate and directly invalidated targets
        """
        # Valid and invalid instances of the previous evaluated shape (if any)
        prev_val_list = set() if filtering_shape is None else filtering_shape.get_valid_targets()
        prev_inv_list = set() if filtering_shape is None else filtering_shape.get_invalid_targets()
        pending_targets = set()
        inv_targets = set()

        self.stats.update_log(''.join(['\n', 'instances retrieval ', shape.get_id(),
                              ": [out-neighbor's (", filtering_shape.get_id(), ')']))
        self.stats.update_log(''.join([' instances: ', str(len(prev_val_list)), ' valid ',
                              str(len(prev_inv_list)), ' invalid]']))

        max_split_number = 256
        max_instances_per_query = 115
        shortest_inst_list = prev_val_list if len(prev_val_list) < len(prev_inv_list) else prev_inv_list

        if prev_val_list == prev_inv_list or \
                len(prev_val_list) == 0 or len(prev_inv_list) == 0 or \
                len(shortest_inst_list) > max_split_number:
            pending_targets = self.extract_targets(shape)
            return pending_targets, inv_targets

        query_template = shape.queriesFilters[filtering_shape.get_id()]
        constraint = query_template['constraint']
        query_template = query_template['query_valid'].get_sparql() if shortest_inst_list == prev_val_list else query_template['query_invalid'].get_sparql()
        separator = ' '

        split_instances = self.__get_formatted_instances(shortest_inst_list, separator, max_instances_per_query)
        query = [query_template.replace('$instances_to_add$', sublist) for sublist in split_instances]

        start = time.time() * 1000.0
        for q in query:
            self.stats.update_log('\nEvaluating target query for ' + shape.id + ':\n' + q)
            bindings = self.endpoint.run_query(q)['results']['bindings']
            for b in bindings:
                instance = b['x']['value']
                cardinality = int(b['cnt']['value'])

                if isinstance(constraint, MinOnlyConstraint):
                    if cardinality < constraint.min:
                        inv_targets.update([(shape.id, instance, True)])
                    else:
                        pending_targets.update([(shape.id, instance, True)])
                elif isinstance(constraint, MaxOnlyConstraint):
                    if cardinality > constraint.max:
                        inv_targets.update([(shape.id, instance, True)])
                    else:
                        pending_targets.update([(shape.id, instance, True)])
        end = time.time() * 1000.0
        self.stats.record_query_exec_time(end - start)
        self.stats.record_query()
        self.stats.update_log('\nelapsed: ' + str(end - start) + ' ms\n')
        self.stats.update_log('\nNumber of pending targets: ' + str(len(pending_targets)))
        self.stats.update_log('\nNumber of invalid targets: ' + str(len(inv_targets)))
        return pending_targets, inv_targets

    def rewrite_constraint_query(self, shape, q, filtering_shape, q_type, use_selective_queries):
        """
        Filters constraint query with targets from 'filtering_shape' (if any).
        The end-user can restrict number of instances allowed for considering a filtering clause.
        Since the length of a query string is restricted by the SPARQL endpoint configuration, the query might be
        divided into several sub-queries, where each subquery contains at most 'max_instances_per_query' (set to 80).

        :param shape: focus shape being evaluated
        :param q: Query object representing the target query of the focus shape
        :param filtering_shape: shape used to filter the instances of the target query
        :param q_type: query type
        :param use_selective_queries: boolean that indicates if the selective validation is enabled
        :return: list with (possibly partitioned) filtered query or original query if no filter was applied
        """
        max_split_number = shape.get_query_split_threshold()
        max_instances_per_query = 80
        prev_val_list = set() if filtering_shape is None else filtering_shape.get_valid_targets()
        prev_inv_list = set() if filtering_shape is None else filtering_shape.get_invalid_targets()
        query_template = q.get_sparql()

        if use_selective_queries and '$inter_shape_type_to_add$' in query_template:
            for var, inter_shape_name in q.get_inter_shape_refs_names().items():
                inter_shape = self.shapes_dict[inter_shape_name]
                inter_shape_triple = ''
                if inter_shape.targetType == 'class':
                    inter_shape_triple = get_target_node_statement(inter_shape.targetQueryNoPref).replace('?x', '?' + var)
                    if inter_shape_triple[-1] != '}':
                        inter_shape_triple += '.'
                query_template = query_template.replace('$inter_shape_type_to_add$', inter_shape_triple)
        else:
            query_template = query_template.replace('$inter_shape_type_to_add$', '')

        if use_selective_queries and \
                filtering_shape is not None and \
                len(prev_val_list) > 0 and len(prev_inv_list) > 0 and \
                len(prev_val_list) <= max_split_number:
            values_clauses = ''
            inter_shape_triples = '\n'
            separator = ''
            split_instances = self.__get_formatted_instances(prev_val_list, separator, max_instances_per_query)
            for c in shape.constraints:
                if c.shapeRef == filtering_shape.get_id() and c.min == 1:
                    obj_var = ' ?' + c.variables[0]
                    values_clauses += 'VALUES' + obj_var + ' {$instances$}\n'
                    if q_type == 'max':
                        focus_var = c.varGenerator.get_focus_node_var()
                        inter_shape_triples += '?' + focus_var + ' ' + c.path + obj_var + '.\n'

            return [query_template.replace(
                        '$filter_clause_to_add$',
                        values_clauses.replace('$instances$', sublist) + inter_shape_triples
                    ) for sublist in split_instances]

        return [query_template.replace('$filter_clause_to_add$', '')]

    @staticmethod
    def __get_formatted_instances(instances, separator, max_list_len):
        """
        Takes instances to be added to the query's filtering clause and returns one or more lists with (in)valid
        instances in two possible formats: (a) using commas for instances in the VALUES clause or (b) using
        spaces as separator for the instances in the FILTER NOT IN clause.

        :param instances: shortest instances list (either the one with valid or invalid targets)
        :param separator: string that contains either a comma or a space character
        :param max_list_len: integer that sets the max number of instances in a query
        :return: set containing all instances with the requested format
        """
        chunks = len(instances) / max_list_len
        n = math.ceil(chunks)
        if chunks > 1:  # Get split list
            instances = tuple(instances)
            inst_count = math.ceil(len(instances) / n)
            split_lists = {instances[i:i + inst_count] for i in range(0, inst_count, inst_count)}
            return {separator.join(subList) for subList in split_lists}
        return {separator.join(instances)}



# -*- coding: utf-8 -*-
__author__ = 'Monica Figuera and Philipp D. Rohde'

import json
import os
from urllib.parse import urlparse

import collections
import rdflib.term
from rdflib import Graph
from itertools import islice

from TravSHACL.core.Shape import Shape
from TravSHACL.utils.VariableGenerator import VariableGenerator
from TravSHACL.constraints.MaxOnlyConstraint import MaxOnlyConstraint
from TravSHACL.constraints.MinMaxConstraint import MinMaxConstraint
from TravSHACL.constraints.MinOnlyConstraint import MinOnlyConstraint

QUERY_TARGET_QUERY = '''SELECT ?query WHERE {{
  <{shape}> a <http://www.w3.org/ns/shacl#NodeShape> ;
      <http://www.w3.org/ns/shacl#targetQuery> ?query .
}}'''

class ShapeParser:
    """Used for parsing shape definitions from files to the internal representation."""

    def __init__(self):
        pass

    def parse_shapes_from_dir(self, path, shape_format, use_selective_queries, max_split_size, order_by_in_queries):
        """
        Parses all files of a certain file extension in the directory.
        It assumes that each file represents one SHACL shape.

        :param path: the path to the directory that stores the shapes files
        :param shape_format: the representation format of the shape definitions
        :param use_selective_queries: indicates whether selective queries are used
        :param max_split_size: maximum number of instances per query
        :param order_by_in_queries: indicates whether to use the ORDER BY clause
        :return: list of Shapes parsed from the files
        """
        file_extension = self.get_file_extension(shape_format)
        files_abs_paths = []

        # r=root, d=directories, f=files
        for r, d, f in os.walk(path):
            for file in f:
                if file_extension in file:
                    files_abs_paths.append(os.path.join(r, file))

        if not files_abs_paths:
            raise FileNotFoundError(path + ' does not contain any shapes of the format ' + shape_format)

        if shape_format == 'JSON':
            return [self.parse_json(
                path=p,
                use_selective_queries=use_selective_queries,
                max_split_size=max_split_size,
                order_by_in_queries=order_by_in_queries
            ) for p in files_abs_paths]
        if shape_format == 'SHACL':
            shapes = []
            [shapes.extend(self.parse_ttl(
                filename=p,
                use_selective_queries=use_selective_queries,
                max_split_size=max_split_size,
                order_by_in_queries=order_by_in_queries
            )) for p in files_abs_paths]
            return shapes
        else:
            print('Unexpected format: ' + shape_format)

    @staticmethod
    def get_file_extension(shape_format):
        if shape_format == 'SHACL':
            return '.ttl'
        else:
            return '.json'  # dot added for convenience

    def parse_json(self, path, use_selective_queries, max_split_size, order_by_in_queries):
        """
        Parses a particular file and converts its content into the internal representation of a SHACL shape.

        :param path: the path to the shapes file
        :param use_selective_queries: indicates whether selective queries are used
        :param max_split_size: maximum number of instances per query
        :param order_by_in_queries: indicates whether to use the ORDER BY clause
        :return: Shape object representing the parsed SHACL shape
        """
        target_query = None
        target_type = None

        file = open(path, 'r')
        obj = json.load(file)
        target_def = obj.get('targetDef')
        name = obj['name']
        id_ = name + '_d1'  # str(i + 1) but there is only one set of conjunctions
        constraints = self.parse_constraints(obj['constraintDef']['conjunctions'], target_def, id_)

        include_sparql_prefixes = self.abbreviated_syntax_used(constraints)
        prefixes = None
        if 'prefix' in obj.keys():
            prefixes = obj['prefix']
        referenced_shapes = self.shape_references(obj['constraintDef']['conjunctions'][0])

        if target_def is not None:
            target_query = target_def['query']

            target_def_copy = target_def.copy()
            del target_def_copy['query']
            target_type = list(target_def_copy.keys())[0]

            if urlparse(target_def[target_type]).netloc != '':  # if the target node is a url, add '<>' to it
                target_def = '<' + target_def[target_type] + '>'
            else:
                target_def = target_def[target_type]

        return Shape(name, target_def, target_type, target_query, constraints, id_, referenced_shapes,
                     use_selective_queries, max_split_size, order_by_in_queries, include_sparql_prefixes, prefixes)

    def parse_ttl(self, filename, use_selective_queries, max_split_size, order_by_in_queries):
        """
        Parses a particular file and converts its content into the internal representation of a SHACL shape.

        :param filename: the path to the shapes file
        :param use_selective_queries: indicates whether selective queries are used
        :param max_split_size: maximum number of instances per query
        :param order_by_in_queries: indicates whether to use the ORDER BY clause
        :return: Shape object representing the parsed SHACL shape
        """
        # TODO: This needs to handle more than one shape per file!
        g_file = Graph()  # create graph instance
        g_file.parse(filename)

        queries = self.get_QUERY()
        shapes = []

        names = [str(row[0]) for row in g_file.query(queries[0])]

        for name in names:
            id_ = name + '_d1'  # str(i + 1) but there is only one set of conjunctions

            # to get the target_ref and target_type
            target_def = None
            target_type = None
            if len(g_file.query(queries[1].format(shape=name))) != 0:
                for res in g_file.query(queries[1].format(shape=name)):
                    target_def = str(res[0])
                    target_type = 'class'
                    break
            elif len(g_file.query(queries[2].format(shape=name))) != 0:
                for res in g_file.query(queries[2].format(shape=name)):
                    target_def = str(res[0])
                    target_type = 'node'
                    break

            target_query = None
            if target_def is not None and target_type == 'class':
                for res in g_file.query(QUERY_TARGET_QUERY.format(shape=name)):
                    target_query = str(res[0])
                if target_query is None:
                    target_query = 'SELECT ?x WHERE { ?x a <' + target_def + '> }'  # come up with a query for this
                    if urlparse(target_def).netloc != '':  # if the target node is a url, add '<>' to it
                        target_def = '<' + target_def + '>'

            cons_dict = self.parse_all_const(g_file, name=name, target_def=target_def, target_type=target_type, query=queries)

            const_array = list(cons_dict.values())  # change the format to an array
            constraints = self.parse_constraints_ttl(const_array, target_def, id_)

            include_sparql_prefixes = self.abbreviated_syntax_used(constraints)
            prefixes = None
            referenced_shapes = self.shape_references(const_array)

            # helps to navigate the shape.__compute_target_queries function
            referenced_shape = {'<' + key + '>': '<' + referenced_shapes[key] + '>'
                                for key in referenced_shapes.keys()
                                if urlparse(referenced_shapes[key]).netloc != ''}

            # to helps to navigate the ShapeSchema.compute_edges function
            if urlparse(name).netloc != '':
                name_ = '<' + name + '>'
            else:
                name_ = name

            shapes.append(Shape(name_, target_def, target_type, target_query, constraints, id_, referenced_shape,
                     use_selective_queries, max_split_size, order_by_in_queries, include_sparql_prefixes, prefixes))

        return shapes

    @staticmethod
    def abbreviated_syntax_used(constraints):
        """
        Run after parsingConstraints.
        Returns false if the constraints' predicates are using absolute paths instead of abbreviated ones
        :param constraints: all shape constraints
        :return: True if prefix notation is used, False otherwise
        """
        for c in constraints:
            if c.path.startswith('<') and c.path.endswith('>'):
                return False
        return True

    @staticmethod
    def shape_references(constraints):
        """
        Gets the shapes referenced by the given constraints.

        :param constraints: the constraints to get the referenced shapes for
        :return: Python dictionary with the referenced shapes and the path referencing the shape
        """
        return {c.get('shape'): c.get('path') for c in constraints if c.get('shape') is not None}

    @staticmethod
    def chunks(datei, SIZE):
        """
        Break down list of input (dictionaries) into individual dictionaries

        :param datei: list of input to be broken down
        :param SIZE: tells the length or size for each output
        :return: inputs divided in len(SIZE)
        """
        it = iter(datei)
        for i in range(0, len(datei), SIZE):
            yield {k: datei[k] for k in islice(it, SIZE)}

    @staticmethod
    def get_QUERY():
        QUERY_SHAPES = '''SELECT DISTINCT ?shape WHERE {
            ?shape a <http://www.w3.org/ns/shacl#NodeShape> .
            }'''

        QUERY_TARGET_1 = '''SELECT ?target WHERE {{
            <{shape}> a <http://www.w3.org/ns/shacl#NodeShape> .
            <{shape}> <http://www.w3.org/ns/shacl#targetClass> ?target .
        }}
        '''

        QUERY_TARGET_2 = '''SELECT ?target WHERE {{
                    <{shape}> a <http://www.w3.org/ns/shacl#NodeShape> .
                    <{shape}> <http://www.w3.org/ns/shacl#targetNode> ?target .
                }}
                '''

        QUERY_CONSTRAINTS = '''SELECT ?constraint WHERE {{
          <{shape}> a <http://www.w3.org/ns/shacl#NodeShape> .
          <{shape}> <http://www.w3.org/ns/shacl#property> ?constraint .
        }}
        '''

        QUERY_CONSTRAINT_DETAILS = '''SELECT ?p ?o WHERE {{
          ?s ?p ?o .
          FILTER( str(?s) = "{constraint}" )
        }}'''

        QUERY_QVS_REF_1 = '''SELECT ?shape_ref WHERE {{
              ?s <http://www.w3.org/ns/shacl#node> ?shape_ref .
              FILTER ( str(?s) = "{qvs}" )
            }}'''

        QUERY_QVS_REF_2 = '''SELECT ?shape_ref WHERE {{
                  ?s <http://www.w3.org/ns/shacl#value> ?shape_ref .
                  FILTER ( str(?s) = "{qvs}" )
                }}'''

        return QUERY_SHAPES, QUERY_TARGET_1, QUERY_TARGET_2, QUERY_CONSTRAINTS, QUERY_CONSTRAINT_DETAILS, QUERY_QVS_REF_1, QUERY_QVS_REF_2

    def get_res(self, filename, name, query):
        """

        :param query: List of queries
        :param name: name of the Shape
        :param filename: shape file in ttl format
        :return: valid response from query execution
        """
        exp_dict = collections.defaultdict(list)
        for constraint in filename.query(query[3].format(shape=name)):
            constraint_id = constraint[0]

            for detail in filename.query(query[4].format(constraint=constraint_id)):

                if isinstance(detail.asdict()['o'], rdflib.term.BNode):
                    qv_type = detail.asdict()['p']
                    qvs = detail.asdict()['o']
                    if len(filename.query(query[5].format(qvs=qvs))) != 0:
                        for shape_ref in filename.query(query[5].format(qvs=qvs)):
                            # dict_1 = [qv_type, ['shape', str(shape_ref.asdict()['shape_ref'])]]
                            dict_1 = [qv_type, str(shape_ref.asdict()['shape_ref'])]
                        exp_dict[str(constraint_id)].append(dict_1.copy())
                    else:
                        for shape_ref in filename.query(query[6].format(qvs=qvs)):
                            dict_1 = [qv_type, ['value', str(shape_ref.asdict()['shape_ref'])]]
                        exp_dict[str(constraint_id)].append(dict_1.copy())
                else:
                    #detail_dict = detail.asdict()
                    dict_2 = [str(detail['p']), str(detail['o'])]
                    exp_dict[str(constraint_id)].append(dict_2.copy())

        return exp_dict

    def parse_all_const(self, filename, name, target_def, target_type, query):
        """

        :param query: List of queries
        :param name: name of the shape
        :param target_type: indicates the target type of the shape, e.g., class or node
        :param target_def: target definition of the shape
        :param filename: file path
        :return: all constraints belonging to the shape
        """
        cons_dict = self.get_res(filename, name, query)
        trav_dict = {}
        exp_dict = {}

        trav_dict['name'] = name
        trav_dict['target_def'] = target_def
        trav_dict['target_type'] = target_type

        for item in self.chunks({i: j for i, j in cons_dict.items()}, 1):
            for dk, dv in item.items():
                if type(dk) is not tuple:

                    trav_dict['min'] = None
                    trav_dict['max'] = None
                    trav_dict['value'] = None
                    trav_dict['path'] = None
                    trav_dict['shape'] = None
                    trav_dict['datatype'] = None
                    trav_dict['negated'] = None

                    for i in dv:
                        if 'path' in str(i[0]).lower():
                            trav_dict['path'] = str(i[1])

                        if 'min' in str(i[0]).lower():
                            trav_dict['min'] = str(i[1])

                        if 'max' in str(i[0]).lower():
                            trav_dict['max'] = str(i[1])

                        if 'datatype' in str(i[0]).lower():
                            trav_dict['datatype'] = str(i[1])

                        if 'valueshape' in str(i[0]).lower():
                            trav_dict['shape'] = str(i[1])
#                            if 'value' in str(i[1][0]).lower():
#                                trav_dict['value'] = str(i[1][1])
#                            if 'shape' in str(i[1][0]).lower():
#                                trav_dict['shape'] = str(i[1][1])

                        if 'not' in str(i[0]).lower():
                            trav_dict['negated'] = str(i[1])

                exp_dict[str(dk)] = trav_dict.copy()

        return exp_dict

    def parse_constraints(self, array, target_def, constraints_id):
        """
        Parses all constraints of a shape.

        :param array: list of constraints belonging to the shape
        :param target_def: the target definition of the shape
        :param constraints_id: suffix for the constraint IDs
        :return: list of constraints in internal constraint representation
        """
        var_generator = VariableGenerator()
        constraints = []
        [constraints.extend(self.parse_constraint(var_generator, array[0][i], constraints_id + '_c' + str(i + 1), target_def)) for i in range(len(array[0]))]
        return constraints

    def parse_constraints_ttl(self, array, target_def, constraints_id):
        """
        Parses all constraints of a shape.

        :param array: list of constraints belonging to the shape
        :param target_def: the target definition of the shape
        :param constraints_id: suffix for the constraint IDs
        :return: list of constraints in internal constraint representation
        """
        var_generator = VariableGenerator()
        constraints = []
        [constraints.extend(self.parse_constraint(var_generator, array[i], constraints_id + '_c' + str(i + 1), target_def)) for i in range(len(array))]
        return constraints

    @staticmethod
    def parse_constraint(var_generator, obj, id_, target_def):
        """
        Parses one constraint to the internal representation.

        :param var_generator: reference to the VariableGenerator instance for variable generation for SPARQL queries
        :param obj: the constraint in its original representation
        :param id_: suffix for the constraint ID
        :param target_def: the target definition of the associated shape
        :return: constraint in internal representation
        """
        min_ = obj.get('min')
        max_ = obj.get('max')
        shape_ref = obj.get('shape')
        datatype = obj.get('datatype')
        value = obj.get('value')
        path = obj.get('path')
        negated = obj.get('negated')

        o_min = None if (min_ is None) else int(min_)
        o_max = None if (max_ is None) else int(max_)
        o_shape_ref = None if (shape_ref is None) else str(shape_ref)
        o_datatype = None if (datatype is None) else str(datatype)
        o_value = None if (value is None) else str(value)
        o_path = None if (path is None) else str(path)
        o_neg = True if (negated is None) else not negated  # True means it is a positive constraint

        if urlparse(path).netloc != '':  # if the predicate is a url, add '<>' to it
            o_path = '<' + path + '>'

        if urlparse(shape_ref).netloc != '' and shape_ref is not None:  # if the shape reference is a url, add '<>' to it
            o_shape_ref = '<' + shape_ref + '>'

        if urlparse(value).netloc != '' and value is not None:  # if the value reference is a url, add '<>' to it
            o_value = '<' + value + '>'

        if urlparse(datatype).netloc != '' and datatype is not None:  # if the data type is a url, add '<>' to it
            o_datatype = '<' + datatype + '>'

        if o_path is not None:
            if o_min is not None:
                if o_max is not None:
                    #return MinMaxConstraint(var_generator, id_, o_path, o_min, o_max, o_neg, o_datatype, o_value, o_shape_ref, target_def)
                    return [MinOnlyConstraint(var_generator, id_, o_path, o_min, o_neg, o_datatype, o_value, o_shape_ref, target_def), MaxOnlyConstraint(var_generator, id_, o_path, o_max, o_neg, o_datatype, o_value, o_shape_ref, target_def)]
                return [MinOnlyConstraint(var_generator, id_, o_path, o_min, o_neg, o_datatype, o_value, o_shape_ref, target_def)]
            if o_max is not None:
                return [MaxOnlyConstraint(var_generator, id_, o_path, o_max, o_neg, o_datatype, o_value, o_shape_ref, target_def)]



#!/usr/bin/env python3
from setuptools import setup

with open('VERSION', 'r', encoding='utf-8') as ver:
    version = ver.read()

with open('README.md', 'r', encoding='utf-8') as fh:
    long_description = fh.read()

setup(
    name='TravSHACL',
    version=version,
    packages=['TravSHACL', 'TravSHACL.constraints', 'TravSHACL.core', 'TravSHACL.rule_based_validation', 'TravSHACL.sparql', 'TravSHACL.utils'],
    license='GNU/GPLv3',
    author='Mónica Figuera, Philipp D. Rohde',
    author_email='philipp.rohde@tib.eu',
    url='https://github.com/SDM-TIB/Trav-SHACL',
    download_url='https://github.com/SDM-TIB/Trav-SHACL/archive/refs/tags/v' + version + '.tar.gz',
    description='A SHACL validator capable of planning the traversal and execution of the validation of a shape schema to detect violations early.',
    long_description=long_description,
    long_description_content_type='text/markdown',
    install_requires=['SPARQLWrapper==2.0.0'],
    python_requires='>=3.7',
    classifiers=[
        'Development Status :: 5 - Production/Stable ',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3 :: Only',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'Operating System :: OS Independent',
        'Intended Audience :: Science/Research'
      ]
)



# -*- coding: utf-8 -*-
from __future__ import annotations

__author__ = 'Monica Figuera and Philipp D. Rohde'

from typing import TYPE_CHECKING

from TravSHACL.constraints.MaxOnlyConstraint import MaxOnlyConstraint

if TYPE_CHECKING:
    from TravSHACL.core.Shape import Shape

from TravSHACL.utils.VariableGenerator import VariableGenerator
from TravSHACL.constraints.Constraint import Constraint
from TravSHACL.core.RulePattern import RulePattern


def get_target_node_statement(target_query):
    """
    Given a target query, this method returns the query body, i.e., the triple patterns.

    :param target_query: complete SPARQL query
    :return: query body of the given SPARQL query
    """
    start = target_query.index('{') + len('{')
    end = target_query.rfind('}')
    return target_query[start:end]


class Query:
    """Internal representation of a SPARQL query."""

    def __init__(self, id_, rule_pattern, sparql, inter_shape_refs=None, max_zero=False):
        """
        Creates a new instance of a SPARQL query.

        :param id_: internal query id
        :param rule_pattern: query rule pattern associated with the query
        :param sparql: actual SPARQL query, i.e., a string complying with the SPARQL protocol
        :param inter_shape_refs: used to note references to other shapes
        :param max_zero: whether this is a max zero query (needs special treatment)
        """
        self.id = id_
        self.rule_pattern = rule_pattern
        self.sparql = sparql
        self.inter_shape_refs = inter_shape_refs
        self.max_zero = max_zero

    def get_id(self):
        return self.id

    def get_rule_pattern(self):
        return self.rule_pattern

    def get_sparql(self):
        return self.sparql

    def get_inter_shape_refs_names(self):
        return self.inter_shape_refs


class QueryGenerator:
    """This class is responsible for generating target and constraint queries."""

    def __init__(self, shape: Shape):
        self.shape = shape

    # TARGET QUERIES #

    def generate_target_query(self, type_, ref_constraint, target_query, include_prefixes, include_order_by):
        """
        Generates a SPARQL target query. It can be a simple query with only one triple pattern or contain
        a VALUES or FILTER NOT IN clause to filter valid or invalid instances directly during the target extraction.

        :param type_: indicates the type of target query to generate
        :param ref_constraint: might hold a list of constraints that refer to this target query
        :param target_query: target query string parsed from the input file
        :param include_prefixes: indicates whether prefixes should be included in the query
        :param include_order_by: indicates whether the ORDER BY clause will be added
        :return: generated query
        """
        if target_query is None:
            return None
        if type_ == 'plain_target':
            return self._plain_target_query(target_query, include_prefixes, include_order_by)
        elif type_ == 'template_FILTER':
            return self._target_query_filter(ref_constraint, target_query, include_prefixes, include_order_by, True), \
                   self._target_query_filter(ref_constraint, target_query, include_prefixes, include_order_by, False)

    def _plain_target_query(self, target_query, include_prefixes, include_order_by):
        """
        Generates a simple target query from the parsed target query.

        :param target_query: target query string parsed from the input file
        :param include_prefixes: indicates whether prefixes should be included in the query
        :param include_order_by: indicates whether the ORDER BY clause will be added
        :return: simple target query
        """
        prefixes = self.shape.get_prefix_string() if include_prefixes else ''
        focus_var = VariableGenerator.get_focus_node_var()
        return ''.join([prefixes,
                        target_query,
                        ' ORDER BY ?' + focus_var if include_order_by else ''])

    def _target_query_filter(self, constraint, target_query, include_prefixes, include_order_by, filter_by_valid=True):
        """
        Generates a target query for filtering based on links to valid or invalid instances of the neighboring shape.

        :param constraint: constraint that refers to this target query
        :param target_query: target query string parsed from the input file
        :param include_prefixes: indicates whether prefixes should be included in the query
        :param include_order_by: indicates whether the ORDER BY clause will be added
        :param: filter_by_valid: indicates whether valid instances will be used for filtering or invalid ones
        :return: target query with VALUES clause
        """
        prefixes = self.shape.get_prefix_string() if include_prefixes else ''
        ref_path = constraint[0].path
        focus_var = VariableGenerator.get_focus_node_var()
        target_node = get_target_node_statement(target_query)
        count = '(COUNT(DISTINCT ?inst) AS ?cnt)' if filter_by_valid else '((COUNT(DISTINCT ?inst2) - COUNT(DISTINCT ?inst)) AS ?cnt)'
        additional_triple_pattern = '  OPTIONAL { ?' + focus_var + ' ' + ref_path + ' ?inst2 . }\n' if not filter_by_valid else ''
        query = ''.join([
            prefixes,
            'SELECT DISTINCT ?' + focus_var + ' ' + count + ' WHERE {\n ',
            target_node + '\n',
            additional_triple_pattern,
            '  OPTIONAL {\n', '    VALUES ?inst { $instances_to_add$ }. \n',
            '    ?' + focus_var + ' ' + ref_path + ' ?inst .\n  }\n',
            '}\n', ' ORDER BY ?' + focus_var if include_order_by else ''
        ])
        return Query(None, None, query)

    # CONSTRAINT QUERIES #

    def generate_query(self, id_, constraints, is_selective, target_query,
                       include_prefixes, include_order_by, subquery=None):
        """
        Generates a constraint query.

        :param id_: internal query/constraint name
        :param constraints: a list of constraints that refer to this constraint query
        :param is_selective: indicates whether selective queries are used
        :param target_query: target query of the shape associated with the constraint
        :param include_prefixes: indicates whether prefixes should be included in the query
        :param include_order_by: indicates whether the ORDER BY clause will be added
        :param subquery: might hold a sub-query to be used in the constraint query
        :return: the generated constraint query
        """
        rp = self.compute_rule_pattern(constraints, id_)
        builder = QueryBuilder(id_, subquery, rp.variables, is_selective, target_query,
                               constraints, include_order_by, self.shape.get_prefix_string())
        for c in constraints:
            builder.build_clause(c)
        return builder.build_query(rp, include_prefixes)

    @staticmethod
    def compute_rule_pattern(constraints, id_):
        """
        Computes the query rule patterns for the given constraints.

        :param constraints: a list of constraints to compute the rule patterns for
        :param id_: the internal constraint query name associated with the constraints
        :return: query rule pattern for the constraints
        """
        body = []
        for c in constraints:
            body = body + c.compute_rule_pattern_body()

        return RulePattern(
            head=(id_, VariableGenerator.get_focus_node_var(), True),
            body=body
        )

    @staticmethod
    def generate_local_subquery(positive_constraints):
        """
        Generates a local sub-query.

        :param positive_constraints: positive constraints
        :return: sub-query including the constraints
        """
        local_pos_constraints = [c for c in positive_constraints if c.get_shape_ref() is None]

        if len(local_pos_constraints) == 0:
            return None  # optional empty

        builder = QueryBuilder('tmp', None, VariableGenerator.get_focus_node_var(), False)

        for c in local_pos_constraints:
            builder.build_clause(c)

        return builder.get_sparql(False, True)


class QueryBuilder:
    """This class is responsible for actually building SPARQL queries for constraints."""

    def __init__(self, id_, subquery, projected_variables, is_selective, target_query=None,
                 constraints=None, include_order_by: bool = False, prefix_string: str = ''):
        """
        Creates a new instance of QueryBuilder.

        :param id_: internal query name
        :param subquery: sub-query to be included, might be None
        :param projected_variables: list of variables to be projected
        :param is_selective: indicates whether selective queries are used
        :param target_query: target query of the shape associated with the constraint
        :param constraints: a list of constraints that refer to the constraint query
        :param include_order_by: indicates whether the ORDER BY clause will be added
        :param prefix_string: prefix string for the queries being built
        """
        self.id = id_
        self.subquery = subquery
        self.projected_variables = projected_variables
        self.filters = []
        self.triples = []

        self.include_selectivity = is_selective
        self.target_query = target_query
        self.constraints = constraints
        self.include_ORDERBY = include_order_by
        self.inter_shape_refs = {}
        self.prefix_string = prefix_string

    def add_triple(self, path, obj):
        """
        Adds a triple pattern to the constraint query.
        No subject is needed as in a constraint query all triple patterns share the same subject.

        :param path: predicate of the triple pattern
        :param obj: object of the triple pattern
        """
        self.triples.append('?' + VariableGenerator.get_focus_node_var() + ' ' + path + ' ' + obj + '.')

    def add_datatype_filter(self, variable, datatype, is_pos):
        """
        Add a filter based on a datatype.

        :param variable: the variable the filter should be added for
        :param datatype: the expected datatype
        :param is_pos: indicated whether this is a positive constraint
        """
        s = 'datatype(?' + variable + ') = ' + datatype
        self.filters.append(s if is_pos else '!(' + s + ')')

    def add_constant_filter(self, variable, constant, is_pos):
        """
        Add a filter for a constant.

        :param variable: the variable the filter should be added for
        :param constant: the constant value the variable should be assigned
        :param is_pos: indicated whether this is a positive constraint
        """
        s = variable + ' = ' + constant
        self.filters.append(s if is_pos else '!(' + s + ')')

    def get_sparql(self, include_prefixes, is_subquery):
        """
        Get the SPARQL query corresponding to the current state of the QueryBuilder.

        :param include_prefixes: indicates whether prefixes should be included in the query
        :param is_subquery: indicates whether the query is considered to be a sub-query
        :return: the SPARQL query as a string
        """
        if is_subquery:
            return self.__get_query(False)  # create subquery

        prefixes = self.prefix_string if include_prefixes else ''
        outer_query_closing_braces = ''.join(['}\n' if self.subquery is not None else '',
                                              '}' if self.get_triple_patterns() != '' and self.subquery is not None else '',
                                              '}' if self.get_triple_patterns() != '' else ''])
        selective_closing_braces = '}}' if self.include_selectivity and self.target_query is not None else ''

        if len(self.constraints) == 1 and isinstance(self.constraints[0], MaxOnlyConstraint) and self.constraints[0].get_shape_ref() is None:
            target_node = ''
            if self.include_selectivity and self.target_query is not None:
                target_node = get_target_node_statement(self.target_query) + '.\n'

            if self.constraints[0].get_value() is not None:
                pred = self.constraints[0].path
                obj = self.constraints[0].get_value()
                return ''.join([prefixes,
                                self.__get_projection_string(),
                                ' WHERE {\n', target_node,
                                '?', VariableGenerator.get_focus_node_var(), ' ', pred, ' ', obj, '.\n}',
                                ' ORDER BY ?' + VariableGenerator.get_focus_node_var() if self.include_ORDERBY else ''])
            else:
                return ''.join([prefixes,
                                self.__get_projection_string(),
                                ' WHERE {\n', target_node,
                                ' OPTIONAL { ', self.triples[0], ' }\n} GROUP BY ?',
                                VariableGenerator.get_focus_node_var(),
                                ' HAVING (COUNT(DISTINCT ?', self.triples[0].rsplit('?', 1)[1][:-1], ') > ', str(self.constraints[0].max), ')',
                                ' ORDER BY ?' + VariableGenerator.get_focus_node_var() if self.include_ORDERBY else ''])

        query = ''.join([prefixes,
                        self.__get_selective(),
                        self.__get_query(include_prefixes),
                        self.subquery if self.subquery is not None else '',
                        outer_query_closing_braces,
                        selective_closing_braces,
                        ' ORDER BY ?' + VariableGenerator.get_focus_node_var() if self.include_ORDERBY else ''])
        return query

    def __get_query(self, include_prefixes):
        """
        Internal method of the QueryBuilder to generate the SPARQL query string.

        :param include_prefixes: indicates whether prefixes should be included in the query
        :return: the SPARQL query as a string
        """
        temp_string = ''
        if include_prefixes:
            if '_pos' in self.id or '_max_' in self.id:
                # add VALUES clause to external query
                temp_string = '$filter_clause_to_add$'

        triple_patterns = self.get_triple_patterns()
        if triple_patterns != '':
            return ''.join([self.__get_projection_string(),
                            ' WHERE {\n',
                            temp_string,
                            triple_patterns,
                            '\n', '{\n' if self.subquery is not None else ''])
        else:
            return ''

    def __get_selective(self):
        """
        Internal method of the QueryBuilder to generate selective constraint queries.

        :return: a SPARQL query including the target definition of the shape associated to the constraint query
        """
        if self.include_selectivity and self.target_query is not None:
            target_node = get_target_node_statement(self.target_query)
            return ''.join([self.__get_projection_string(),
                            ' WHERE {\n',
                            target_node + '. {\n'])
        return ''

    def __get_projection_string(self):
        """
        Internal method of the QueryBuilder for generating the SELECT clause.

        :return: the SELECT clause of the query to be built
        """
        return 'SELECT DISTINCT ' + ' '.join(['?' + v for v in self.projected_variables])

    def get_triple_patterns(self):
        """
        Gets all triple patterns belonging to the query that is currently being built.

        :return: all triple patterns of the query
        """
        triple_string = '\n'.join(self.triples)

        if len(self.filters) == 0:
            return triple_string

        return triple_string + self.generate_filter_string()

    def generate_filter_string(self):
        """
        Generates the filter string for the filters to be applied to the query.

        :return: a string for the filters to be applied
        """
        if len(self.filters) == 0:
            return ''

        return '\nFILTER(\n' + (self.filters[0] if len(self.filters) == 1 else ' &&\n'.join(self.filters)) + ')'

    def add_cardinality_filter(self, variables):
        """
        Adds a filter on the cardinality of a variable.

        :param variables: list of variables to add cardinality filters for
        """
        for i in range(0, len(variables)):
            for j in range(i + 1, len(variables)):
                self.filters.append('?' + variables[i] + ' != ?' + variables[j])

    def build_clause(self, c):
        """
        Adds the necessary information to the QueryBuilder for a given constraint.

        :param c: the constraint to be included in the query
        """
        variables = c.get_variables()

        if isinstance(c, Constraint):
            path = c.path

            if c.get_value() is not None:  # if there is fixed value for the object
                self.add_triple(path, c.get_value())
                return

            for v in variables:
                if c.get_shape_ref() is not None:  # if there is an existing reference to another shape
                    self.inter_shape_refs[v] = c.get_shape_ref()
                    self.triples.append('\n$inter_shape_type_to_add$')

                self.add_triple(path, '?' + v)

        if c.get_value() is not None:
            self.add_constant_filter(
                variables.iterator().next(),
                c.get_value().get(),
                c.get_is_pos()
            )

        if c.get_datatype() is not None:
            for v in variables:
                self.add_datatype_filter(v, c.get_datatype(), c.get_is_pos())

        if len(variables) > 1:
            self.add_cardinality_filter(variables)

    def build_query(self, rule_pattern, include_prefixes):
        """
        Generate a Query object based on the current state of the QueryBuilder.

        :param rule_pattern: the query rule pattern associated with the query
        :param include_prefixes: indicates whether prefixes should be included in the query
        """
        max_zero_query = False
        for c in self.constraints:
            if c.max == 0:
                max_zero_query = True

        return Query(
            self.id,
            rule_pattern,
            self.get_sparql(include_prefixes, False),
            self.inter_shape_refs,
            max_zero_query
        )



# -*- coding: utf-8 -*-
__author__ = 'Philipp D. Rohde'

import os

from TravSHACL.core.GraphTraversal import GraphTraversal
from TravSHACL.core.ShapeSchema import ShapeSchema


def create_output_dir(output_dir):
    """
    Creates the output directory if it does not exist.

    :param output_dir: the output directory to be used
    """
    path = os.getcwd()
    os.makedirs(path + '/' + output_dir, exist_ok=True)


def parse_heuristics(input_):
    """
    Parses the heuristics from the arguments passed to Trav-SHACL.

    :param input_: the heuristics argument
    :return: Python dictionary with the heuristics to be used for the evaluation
    """
    heuristics = {}
    if 'TARGET' in input_:
        heuristics['target'] = True
    else:
        heuristics['target'] = False

    if 'IN' in input_:
        heuristics['degree'] = 'in'
    elif 'OUT' in input_:
        heuristics['degree'] = 'out'
    else:
        heuristics['degree'] = False

    if 'SMALL' in input_:
        heuristics['properties'] = 'small'
    elif 'BIG' in input_:
        heuristics['properties'] = 'big'
    else:
        heuristics['properties'] = False

    return heuristics


def eval_shape_schema(args):
    """
    Evaluates a SHACL shape schema against a SPARQL endpoint.

    :param args: command line arguments passed to Trav-SHACL
    """
    create_output_dir(args.outputDir)
    shape_schema = ShapeSchema(
        schema_dir=args.d,
        schema_format='JSON' if args.json else 'SHACL',
        endpoint_url=args.endpoint,
        graph_traversal=GraphTraversal.BFS if args.graphTraversal == 'BFS' else GraphTraversal.DFS,
        heuristics=parse_heuristics(args.heuristics),
        use_selective_queries=args.selective,
        max_split_size=args.m,
        output_dir=args.outputDir,
        order_by_in_queries=args.orderby,
        save_outputs=args.outputs
    )

    report = shape_schema.validate()  # run the evaluation of the SHACL constraints over the specified endpoint
    if not args.outputs:
        print('Report:', report)
